#ifndef COGS_INCLUDE_PHONGLIGHTING
#define COGS_INCLUDE_PHONGLIGHTING

#ifndef COGS_MAX_LIGHTS
#define COGS_MAX_LIGHTS 8
#endif


#ifdef COGS_FORWARD_SHADOWS_ENABLED
#include "Engine/CascadedShadowMaps.hlsl"
#include "Engine/PointLightShadowMaps.hlsl"
#endif

TextureCube environmentSky;
SamplerState environmentSkySampler;

TextureCube environmentRadiance;
SamplerState environmentRadianceSampler;

TextureCube environmentIrradiance;
SamplerState environmentIrradianceSampler;

TextureCube ambientIrradiance;
SamplerState ambientIrradianceSampler;

cbuffer LightBuffer
{
  float4 lightPositions[COGS_MAX_LIGHTS];
  float4 lightDirections[COGS_MAX_LIGHTS];
  float4 lightColorIntensity[COGS_MAX_LIGHTS];
  float4 lightParameters[COGS_MAX_LIGHTS];
  uint4 numLights;
  float4 eyePosition;
  float4 fogColor;
  float fogDistance;
  float fogAmount;
  int fogEnabled;
  float ambientIntensity;
  float3 ambientColor;
  float environmentBrightness;
  float skyMultiplier;
  int seaFlags;
  int flags;  // 1: has skydome, 2: has radiance, 3: has irradiance, 4: has subsea radiance.
};

float3 PhongLighting(uint i, float3 diffuseColor, float3 normal, float3 worldPosition, float3 specularColor, float specularPower)
{
  float4 lightPosition = lightPositions[i];

  float3 offset = worldPosition - lightPosition.xyz;
  float3 lightDir = lightDirections[i].w == 0.f ? lightDirections[i].xyz : normalize(offset);

  float distance = length(offset);
  float radius = lightParameters[i].a;

  float attenuation = lightDirections[i].w == 0 ? 1 : saturate(1.0 - (distance * distance) / (radius * radius));
  attenuation *= attenuation;

  float lightIntensity = attenuation * lightColorIntensity[i].a;
  float diffuseIntensity = saturate(dot(normal, -lightDir));

  float3 color = diffuseColor.rgb * lightColorIntensity[i].rgb * diffuseIntensity * lightIntensity;

  if (max(max(specularColor.r, specularColor.g), specularColor.b) > 0) {
    float3 eyeDir = -normalize(eyePosition.xyz - worldPosition);
    float3 halfDir = normalize(lightDir + eyeDir);
    float specularIntensity = pow(saturate(dot(normal, -halfDir)), specularPower);

    color += specularColor * lightColorIntensity[i].rgb * specularIntensity * lightIntensity;
  }

  return color;
}

struct SurfaceLighting
{
  float4 Color;
  float3 Normal;
  float3 Specular;
  float3 ViewSpacePosition;
  float SpecularPower;
  float3 Emissive;
  int LightModel;
  float4 WorldPosition;
};

struct VolumeLighting
{
  float4 Color;
  float3 Normal;
  float3 Specular;
  float3 ViewSpacePosition;
  float SpecularPower;
  float3 Emissive;
  int LightModel;
  float4 WorldPositionA;
  float4 WorldPositionB;
};

// Applies simple fog color.
float3 applySimpleFog(in SurfaceLighting surface, float3 inputColor)
{
  const float distance = length(surface.ViewSpacePosition);

  float3 viewDir = -normalize(eyePosition.xyz - surface.WorldPosition.xyz);

#ifdef COGS_IMAGE_BASED_LIGHTING
  float3 fogCol = fogColor.rgb * (environmentBrightness * environmentRadiance.SampleLevel(environmentRadianceSampler, (1.0 / distance) * viewDir.xzy, 3.0).rgb);
#else
  float3 fogCol = fogColor.rgb;
#endif
  if ((fogEnabled & 3) == 2) {

    fogCol = environmentBrightness * environmentIrradiance.SampleLevel(environmentIrradianceSampler, (1.0 / distance) * viewDir.xzy, 0.0).rgb;
  }

  float t;
  if ((fogEnabled & 4) != 0) {
    // Total occulsion behaviour
    // 1 - e^(-5) is pretty flat and close to 1. We make it so that
    // the function is exactly one here so we can cull geometry
    // further away. Useful for dense fog.
    // 1/(1-e^{-5})=1.0068.
    float fogDensity = 5.0 / fogDistance;
    t = min(1.0, 1.0068*(1.0 - exp(-distance * fogDensity)));
  } else {
    // classic behaviour
    const float maxFogAmount = 1.0;
    float fogDensity = 1.0 / fogDistance;
    float fogAmount = 1.0 - exp(-distance * fogDensity);
    t = clamp(fogAmount, 0.0, maxFogAmount);
  }

  return lerp(inputColor, fogCol, t);
}

float3 computeFog(in SurfaceLighting surface, float3 inputColor)
{
  return applySimpleFog(surface, inputColor);
}


float4 DoLighting(in VolumeLighting volume, float4 screenPos)
{
  float4 outColor = float4(0, 0, 0, 0);

  if ((volume.LightModel & COGS_LIGHT_MASK_PHONG) == COGS_LIGHT_MASK_PHONG) {
    float3 color = float3(0, 0, 0);

    for (uint i = 0; i < COGS_MAX_LIGHTS; ++i) {
      float shadowFactor = 1;

#ifdef COGS_FORWARD_SHADOWS_ENABLED
      if ( ((volume.LightModel & COGS_LIGHT_MASK_SHADOW_RECEIVER) == COGS_LIGHT_MASK_SHADOW_RECEIVER) && (lightPositions[i].w == 0)) {
#if COGS_NUM_DIRECTIONAL_SHADOWS > 0
        shadowFactor = calculateShadowContribution(-volume.ViewSpacePosition.z,
                                                   volume.WorldPositionA,
                                                   volume.WorldPositionB,
                                                   lightDirections[i].xyz,
                                                   i,
                                                   screenPos);
#endif
      }
#endif

      float shadowIntensityOffset = lightParameters[i].r;
      shadowFactor = min(shadowFactor + shadowIntensityOffset, 1);

      color += shadowFactor * PhongLighting(i, volume.Color.rgb, volume.Normal, volume.WorldPositionA.xyz, volume.Specular, volume.SpecularPower);
    }

    color += volume.Color.rgb * ambientColor * ambientIntensity;

    outColor = float4(color + volume.Emissive.rgb, volume.Color.a);
  } else {
    outColor = float4(volume.Color.rgb + volume.Emissive, volume.Color.a);
  }

  return outColor;
}


float4 DoLighting(in SurfaceLighting surface, float4 screenPos, float4 indirectLighting = float4(0, 0, 0, 1))
{
  float4 outColor = float4(0, 0, 0, 0);

  if ((surface.LightModel & COGS_LIGHT_MASK_PHONG) == COGS_LIGHT_MASK_PHONG) {
    float3 color = surface.Color.rgb * indirectLighting.rgb;

    for (uint i = 0; i < COGS_MAX_LIGHTS; ++i) {
      float4 lightPosition = lightPositions[i];
      float shadowFactor = indirectLighting.a;

#ifdef COGS_FORWARD_SHADOWS_ENABLED
      if (((surface.LightModel & COGS_LIGHT_MASK_SHADOW_RECEIVER) == COGS_LIGHT_MASK_SHADOW_RECEIVER) && (lightPositions[i].w == 0)) {
#if COGS_NUM_DIRECTIONAL_SHADOWS > 0
        shadowFactor = indirectLighting.a * calculateShadowContribution(surface.WorldPosition, surface.Normal, lightDirections[i].xyz, i, screenPos);
#endif
      } else {
#if COGS_NUM_POINT_SHADOWS > 0
        shadowFactor = indirectLighting.a * calculatePointLightShadow(lightPosition, surface.WorldPosition.xyz, surface.Normal, i);
#endif
      }
#endif

      float shadowIntensityOffset = lightParameters[i].r;
      shadowFactor = min(shadowFactor + shadowIntensityOffset, 1);

#ifdef COGS_IMAGE_BASED_LIGHTING
      if (i == 0) {
        uint width, height, levels;
        environmentRadiance.GetDimensions(0, width, height, levels);

        float3 eyeDir = -normalize(eyePosition.xyz - surface.WorldPosition.xyz);
        float3 refDir = reflect(eyeDir, surface.Normal);

        float l = levels * 0.001*max(0, 80 - surface.SpecularPower);

        color += shadowFactor * environmentBrightness * (
          surface.Color.rgb * environmentIrradiance.SampleLevel(environmentIrradianceSampler, surface.Normal.xzy, 0.0).rgb +
          surface.Specular.rgb * environmentRadiance.SampleLevel(environmentRadianceSampler, refDir.xzy, l).rgb);

        continue;
      }
#endif
      color += shadowFactor * PhongLighting(i, surface.Color.rgb, surface.Normal, surface.WorldPosition.xyz, surface.Specular, surface.SpecularPower);
        
    }

#ifdef COGS_IMAGE_BASED_LIGHTING
    color += surface.Color.rgb * ambientIntensity * ambientIrradiance.SampleLevel(ambientIrradianceSampler, surface.Normal.xzy, 0.0).rgb;
#else
    color += surface.Color.rgb * ambientColor * ambientIntensity;
#endif

    outColor = float4(color + surface.Emissive.rgb, surface.Color.a);
  } else if ((surface.LightModel & 15) == 0) {
    outColor = float4(surface.Color.rgb + surface.Emissive, surface.Color.a);
  }
  
  if ((surface.LightModel & COGS_LIGHT_MASK_INDIRECT) == COGS_LIGHT_MASK_INDIRECT) {
    float m = 0.5*max(max(surface.Color.r, surface.Color.g), surface.Color.b);

    outColor += float4(indirectLighting.rgb * ((float3)m + surface.Color.rgb) + indirectLighting.a*(surface.Color.rgb + surface.Emissive), surface.Color.a);
  }
  
  if (fogEnabled && fogAmount != 0) {
    outColor.rgb = lerp(outColor.rgb, computeFog(surface, outColor.rgb), fogAmount);
  }

  return outColor;
}


#endif
