#version 100
#extension GL_EXT_shader_texture_lod : enable
precision highp float;
precision highp int;

varying vec2 NormalizedCoords;

uniform sampler2D aoTexture;
uniform sampler2D colorTexture;

uniform float blur_threshold;
uniform vec2 viewportSize;

uniform float exposure;

#if defined(TONEMAP_FILMIC)
// Filmic tonemapper (aka Uncharted2)
//
// http://filmicworlds.com/blog/filmic-tonemapping-operators/

vec3 filmicToneMapOperator(vec3 c)
{
    return ((c*(0.15*c+0.10*0.50)+0.20*0.02)/(c*(0.15*c+0.50)+0.20*0.30))-0.02/0.30;
}

vec3 tonemap(vec3 c)
{
    // Exposure should be divided by 3 to match other curves
    return (1.0/filmicToneMapOperator(vec3(11.2)))*filmicToneMapOperator((1.0/3.0) * 2.0 * exposure * c.rgb);
}

#elif defined(COGS_TONEMAP_ACES_TRIBAND)
// ACES tri-band tonemapper
//
// https://github.com/TheRealMJP/BakingLab/blob/master/BakingLab/ACES.hlsl

vec3 tonemap(vec3 x)
{
    const mat3 ACESInputMat = mat3(0.59719, 0.07600, 0.02840,   // col 0
                                   0.35458, 0.90834, 0.13383,   // col 1 ..
                                   0.04823, 0.01566, 0.83777);

    const mat3 ACESOutputMat = mat3( 1.60475, -0.10208, -0.00327, 
                                    -0.53108,  1.10813, -0.07276,  
                                    -0.07367, -0.00605, 1.07602);
    vec3 t0 = ACESInputMat * 2.0 * exposure * x;
    vec3 t1 = t0 * (           t0 + 0.0245786) - 0.000090537;
    vec3 t2 = t0 * (0.983729 * t0 + 0.4329510) + 0.238081;
    vec3 t3 = t1/t2;
    vec3 t4 = ACESOutputMat * t3;
    return t4;
}

#elif defined(COGS_TONEMAP_ACES_LUMINANCE)
// ACES luminance-only
//
//https://knarkowicz.wordpress.com/2016/01/06/aces-filmic-tone-mapping-curve/

vec3 tonemap(vec3 x)
{
    x = exposure * x;
    const float a = 2.51;
    const float b =0.03;
    const float c =2.43;
    const float d =0.59;
    const float e =0.14;
    return ((x*(a*x+b))/(x*(c*x+d)+e));
}
#endif


void main()
{
    vec2 s = 1.0/viewportSize;

    vec4 ctr = texture2D(aoTexture, NormalizedCoords);

    vec3 n = 2.0*ctr.rgb - vec3(1.0);

    vec4 color = texture2D(colorTexture, NormalizedCoords);
    if(ctr.x + ctr.y + ctr.z != 0.0) {
        float sum = 0.0;
        float wgt = 0.0;
        float avg = 0.0;
        int hits = 0;
        for(int j=-1; j<3; j++) {
            for(int i=-1; i<3; i++) {

                vec4 val = texture2D(aoTexture, NormalizedCoords + vec2(i,j)*s);

                vec3 m = 2.0*val.rgb - vec3(1.0);

                if(blur_threshold < dot(m,n)) {
                    sum += val.a;
                    wgt += 1.0;
                    hits++;
                }
                avg += (1.0/16.0)*val.a;
            }
        }
        float o = sum/wgt;

        if(hits <= 2 ||  avg + 0.5 < o) {
            o = avg;
        }

        color.rgb = o*color.rgb;
    }
    color.rgb = tonemap(color.rgb);

    // Gamma correction
    color.rgb = pow(abs(color.rgb), vec3(1.0/2.2));
    color.a = 1.0;

#if defined(COGS_LUMINANCE_IN_ALPHA)
    float luma = dot(color.rgb, vec3(0.299, 0.587, 0.114));
    color.a = luma;
#endif

#if 0
    // Gamma test, top and bottom half of screens should match.
    // One half is value 0.5 (50% energy) other half is 50-50 mix of 0.0 and 1.0 (i.e. 50% energy)
    color.rgb = vec3(0.5);
    if(NormalizedCoords.y < 0.5) {
        vec2 pp = floor((viewportSize*NormalizedCoords.xy)/1.0);
        color.rgb = vec3(floor(mod(pp.x+pp.y, 2.0)));
    }
#endif


    gl_FragColor = vec4(color.rgb, color.a);
}

