#ifndef COGS_INCLUDE_STANDARDLIGHTING
#define COGS_INCLUDE_STANDARDLIGHTING

#include "Engine/PhongLighting.es20.glsl"

#define COGS_PI 3.141592653589793238462643383279502884

vec3 GetEnvironmentReflection(samplerCube tex, vec3 direction, float level)
{
  vec3 reflection = textureCube(tex, direction.xzy, level).rgb;
  return environmentBrightness() * reflection;
}

float SchlickBeckmann(float NdotL, float NdotV, float k)
{
  float gL = NdotL / (NdotL * (1.0 - k) + k);
  float gV = NdotV / (NdotV * (1.0 - k) + k);

  return gL * gV;
}

float SchlickBeckmannVisibility(float NdotL, float NdotV, float roughness)
{
  float k = (roughness * roughness) * sqrt(2.0 / COGS_PI);

  return SchlickBeckmann(NdotL, NdotV, k);
}

float SmithGGXVisibility(float NdotL, float NdotV, float roughness)
{
  float k = (roughness * roughness) / 2.0;

  return SchlickBeckmann(NdotL, NdotV, k);
}

float KelemenVisibility(float NdotL, float NdotV, float VdotH)
{
  return (NdotL * NdotV) / (VdotH * VdotH);
}

float GGXNormalDistribution(float NdotH, float roughness)
{
  roughness = clamp(roughness, 0.01, 1.0);
  float a = roughness * roughness;
  float a2 = a * a;
  float d = NdotH * NdotH * (a2 - 1.0) + 1.0;

  d = max(0.0000001, d);

  return a2 / (COGS_PI * (d * d));
}

vec3 SchlickFresnel(vec3 F0, float hl)
{
  float t = pow(max(0.0, 1.0 - hl), 5.0);

  return F0 + (1.0 - F0) * t;
}

vec3 StandardLighting(vec3 lightRadiance, vec3 normal, vec3 viewDir, vec3 lightDir, vec3 specularColor, vec3 diffuseColor, float roughness, float reflectivity, float occlusion)
{
  vec3 halfDir = normalize(viewDir + lightDir);
  float VdotH = clamp(dot(viewDir, halfDir), 0.0, 1.0);
  float NdotL = clamp(dot(normal, lightDir), 0.0, 1.0);
  float NdotV = clamp(dot(normal, viewDir), 0.0, 1.0);
  float NdotH = clamp(dot(normal, halfDir), 0.0, 1.0);

  vec3 F = SchlickFresnel(specularColor, VdotH);
  float V = SmithGGXVisibility(NdotL, NdotV, roughness);
  float D = GGXNormalDistribution(NdotH, roughness);

  vec3 specular = max(vec3(0), NdotL * (D * V * F) / (4.0 * NdotL * NdotH));
  float diffuse = clamp(NdotL / COGS_PI, 0.0, 1.0);

  return occlusion * diffuse * diffuseColor * lightRadiance +
    specular * lightRadiance;
}

vec3 StandardAmbient(vec3 normal, vec3 viewDir, vec3 specularColor, vec3 diffuseColor, float roughness, float reflectivity, float occlusion)
{

#ifdef COGS_IMAGE_BASED_LIGHTING
  return occlusion * diffuseColor * ambientColor.rgb * textureCube(ambientIrradiance, normal.xzy, 0.0).rgb;
#endif

  float NdotV = clamp(dot(normal, viewDir), 0.0, 1.0);

  vec3 direction = reflect(-viewDir, normal);
  vec3 radiance = GetEnvironmentReflection(environmentRadiance, direction, roughness * environmentRadianceMips());
  vec3 irradiance = GetEnvironmentReflection(environmentIrradiance, normal, roughness * environmentIrradianceMips());

  return occlusion * irradiance * diffuseColor +
         radiance * mix(specularColor, vec3(clamp(1.0 - roughness + reflectivity, 0.0, 1.0)), pow(max(0.0, 1.0 - NdotV), 5.0));
}

#endif
