#ifndef COGS_INCLUDE_STANDARDLIGHTING
#define COGS_INCLUDE_STANDARDLIGHTING

#include "Engine/PhongLighting.hlsl"

#define COGS_PI 3.141592653589793238462643383279502884

float3 GetEnvironmentReflection(TextureCube tex, float3 direction, float roughness = 0)
{
  uint width, height;
  uint levels;
  tex.GetDimensions(0, width, height, levels);

  float3 reflection = tex.SampleLevel(environmentRadianceSampler, direction.xzy, roughness * levels).rgb;

  return environmentBrightness * reflection;
}

float SchlickBeckmann(float NdotL, float NdotV, float k)
{
  float gL = NdotL / (NdotL * (1 - k) + k);
  float gV = NdotV / (NdotV * (1 - k) + k);

  return gL * gV;
}

float SchlickBeckmannVisibility(float NdotL, float NdotV, float roughness)
{
  float k = (roughness * roughness) * sqrt(2 / COGS_PI);

  return SchlickBeckmann(NdotL, NdotV, k);
}

float SmithGGXVisibility(float NdotL, float NdotV, float roughness)
{
  float k = (roughness * roughness) / 2;

  return SchlickBeckmann(NdotL, NdotV, k);
}

float KelemenVisibility(float NdotL, float NdotV, float VdotH)
{
  return (NdotL * NdotV) / (VdotH * VdotH);
}

float GGXNormalDistribution(float NdotH, float roughness)
{
  roughness = clamp(roughness, 0.01, 1);
  float a = roughness * roughness;
  float a2 = a * a;
  float d = NdotH * NdotH * (a2 - 1.0) + 1.0;

  d = max(0.0000001, d);

  return a2 / (COGS_PI * (d * d));
}

float3 SchlickFresnel(float3 F0, float hl)
{
  float t = pow(1 - hl, 5);

  return F0 + (1.0 - F0) * t;
}

float3 StandardLighting(float3 lightRadiance, float3 normal, float3 viewDir, float3 lightDir, float3 specularColor, float3 diffuseColor, float roughness, float reflectivity, float occlusion)
{
  float3 halfDir = normalize(viewDir + lightDir);
  float VdotH = saturate(dot(viewDir, halfDir));
  float NdotL = saturate(dot(normal, lightDir));
  float NdotV = saturate(dot(normal, viewDir));
  float NdotH = saturate(dot(normal, halfDir));

  float3 F = SchlickFresnel(specularColor, VdotH);
  float V = SmithGGXVisibility(NdotL, NdotV, roughness);
  float D = GGXNormalDistribution(NdotH, roughness);

  float3 specular = max(0, NdotL * (D * V * F) / (4 * NdotL * NdotH));
  float diffuse = saturate(NdotL / COGS_PI);

  return occlusion * diffuse * diffuseColor * lightRadiance +
    specular * lightRadiance;
}

float3 StandardAmbient(float3 normal, float3 viewDir, float3 specularColor, float3 diffuseColor, float roughness, float reflectivity, float occlusion)
{

#ifdef COGS_IMAGE_BASED_LIGHTING
  return occlusion * diffuseColor * ambientIntensity * ambientIrradiance.SampleLevel(ambientIrradianceSampler, normal.xzy, 0.0).rgb;
#endif

  float NdotV = saturate(dot(normal, viewDir));

  float3 direction = reflect(-viewDir, normal);
  float3 radiance = GetEnvironmentReflection(environmentRadiance, direction, roughness);
  float3 irradiance = GetEnvironmentReflection(environmentIrradiance, normal, roughness);

  return occlusion * irradiance * diffuseColor +
         radiance * lerp(specularColor, saturate(1 - roughness + reflectivity), pow(1 - NdotV, 5));
}

#endif
